import 'package:car_app/exception/server.dart';
import 'package:car_app/model/calendar_event.dart';
import 'package:car_app/model/user.dart';
import 'package:car_app/service/interface/authentication.dart';
import 'package:car_app/service/interface/calendar_event.dart';
import 'package:car_app/service/interface/logging.dart';
import 'package:car_app/service/interface/team.dart';
import 'package:car_app/view/defaults.dart';
import 'package:car_app/view/routes_titles.dart';
import 'package:flutter/material.dart';
import 'package:intl/intl.dart';
import 'package:provider/provider.dart';

class CalendarEventView extends StatefulWidget {

  final CalendarEvent? initialEvent;
  final DateTime? initialStartTime;

  const CalendarEventView({super.key, this.initialEvent, this.initialStartTime});

  @override
  // ignore: no_logic_in_create_state
  State<StatefulWidget> createState() => _CalendarEventViewState(
    initialEvent: initialEvent,
    initialStartTime: initialStartTime
  );
}

class _CalendarEventViewState extends State<CalendarEventView> {
  final CalendarEvent? initialEvent;
  final DateTime? initialStartTime;
  final DateFormat standardFormat = DateFormat("dd.MM.yyyy HH:mm");
  final TextEditingController _titleController = TextEditingController();
  final TextEditingController _descriptionController = TextEditingController();
  final List<DropdownMenuItem<String>> _teams = [];
  final TextEditingController _startTimeController = TextEditingController();
  final TextEditingController _endTimeController = TextEditingController();
  final _formKey = GlobalKey<FormState>();

  late final CalendarEventService calendarEventService;
  late final LoggingService loggingService;

  String? _selectedTeam;
  late DateTime _startTime; // To avoid parsing to string and back
  late DateTime _endTime;
  User? _currentUser; // is set in initState

  bool get _isEventFromUser {
    if(initialEvent == null) {
      return true;
    }
    
    if(_currentUser == null) {
      return false;
    }

    return initialEvent!.user.id == _currentUser!.id;
  }

  _CalendarEventViewState({required this.initialEvent, required this.initialStartTime});

  void setStartTime(DateTime time) {
    setState(() {
      _startTime = time;
      _startTimeController.text = standardFormat.format(time);
    });
  }

  void setEndTime(DateTime time) {
    setState(() {
      _endTime = time;
      _endTimeController.text = standardFormat.format(time);
    });
  }

  Future<DateTime?> _selectDateTime(DateTime initial) async {
    var selectedDate = await showDatePicker(
      context: context,
      firstDate: DateTime(1999, 01, 01),
      lastDate: DateTime(2100, 12, 31),
      initialDate: initial,
      initialEntryMode: DatePickerEntryMode.calendarOnly,
      cancelText: "Cancel",
      confirmText: "Confirm",
    );
    
    if (selectedDate != null && mounted) {
      var selectedTime = await showTimePicker(
        context: context,
        initialTime: TimeOfDay.fromDateTime(initial),
        cancelText: "Cancel",
        confirmText: "Confirm",
        builder: (context, child) {
          return MediaQuery(
            data: MediaQuery.of(context).copyWith(alwaysUse24HourFormat: true),
            child: child!,
          );
        },
      );

      if (selectedTime != null) {
        return selectedDate.add(Duration(hours: selectedTime.hour, minutes: selectedTime.minute));
      }
    }

    return null;
  }

  @override
  void initState() {
    calendarEventService = Provider.of(context, listen: false);
    loggingService = Provider.of(context, listen: false);
    var authenticationService = Provider.of<AuthenticationService>(context, listen: false);
    var teamService = Provider.of<TeamService>(context, listen: false);

    _titleController.text = initialEvent?.title ?? "";

    _descriptionController.text = initialEvent?.description ?? "";
    _selectedTeam = initialEvent?.teamId;

    var startTime = initialEvent?.start;
    if(startTime != null) {
      setStartTime(startTime);
    }
    else if(initialStartTime != null) {
      setStartTime(initialStartTime!); 
    }
    else {
      setStartTime(DateTime.now());
    }

    var endTime = initialEvent?.end;
    if(endTime != null) {
      setEndTime(endTime);
    }
    else {
      setEndTime(_startTime.add(const Duration(hours: 1)));
    }

    authenticationService.isLoggedIn().then((isLoggedIn) {
      if(isLoggedIn) {
        authenticationService.user.then((user) {
          _currentUser = user;

          if(_titleController.text.isEmpty) {
            setState(() {
              _titleController.text = initialEvent?.title ?? "Booking for ${user.name}";
            });
          }
        });

        teamService.getTeams().forEach((element) {
          setState(() {
            _teams.add(DropdownMenuItem(
              value: element.teamId,
              child: Text(element.teamName),
            ));
          });
        });
      }
    });

    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: DefaultAppBar(
        context,
        titleText: CalendarEventRouting.title,
      ),
      body: Padding(
        padding: const EdgeInsets.all(20),
        child: Column(
          children: [
            SingleChildScrollView(
              child: Form(
                key: _formKey,
                child: Column(
                  children: [
                    TextFormField(
                      decoration: const InputDecoration(
                        border: OutlineInputBorder(),
                        label: Text("Title"),
                      ),
                      controller: _titleController,
                      validator: (value) {
                        if (value == null || value.isEmpty) {
                          return "Please enter a title";
                        }
                        return null;
                      },
                    ),

                    const SizedBox(height: 10),

                    TextField(
                      decoration: const InputDecoration(
                        border: OutlineInputBorder(),
                        label: Text("Description"),
                      ),
                      controller: _descriptionController,
                    ),

                    const SizedBox(height: 10),

                    DropdownButtonFormField(
                      items: _teams,
                      isExpanded: true,
                      value: _selectedTeam,
                      decoration: const InputDecoration(
                        border: OutlineInputBorder(),
                        label: Text("Team"),
                      ),
                      onChanged: (value) {
                        setState(() {
                          _selectedTeam = value;
                        });
                      },
                      validator: (value) {
                        if (value == null || value.isEmpty) {
                          return "Please select a team";
                        }

                        return null;
                      },
                    ),

                    const SizedBox(height: 10),

                    Row(
                      crossAxisAlignment: CrossAxisAlignment.start, // prevents other textfield to be centered in case of an error
                      children: [
                        Expanded(
                          child: TextField(
                            decoration: const InputDecoration(
                              border: OutlineInputBorder(),
                              label: Text("Start"),
                            ),
                            controller: _startTimeController,
                            onTap: () async {
                              var startTime = await _selectDateTime(_startTime);
                              if (startTime != null) {
                                setStartTime(startTime);
                              }
                            },
                            readOnly: true,
                          ),
                        ),

                        const SizedBox(width: 10),

                        Expanded(
                          child: TextFormField(
                            decoration: const InputDecoration(
                              border: OutlineInputBorder(),
                              label: Text("End"),
                            ),
                            controller: _endTimeController,
                            onTap: () async {
                              var endTime = await _selectDateTime(_endTime);
                              if (endTime != null) {
                                setEndTime(endTime);
                              }
                            },
                            readOnly: true,
                            validator: (_) {
                              if (!(_endTime.isAfter(_startTime) || _endTime.isAtSameMomentAs(_startTime))) {
                                return "End must be after start";
                              }
                              return null;
                            }
                          ),
                        ),
                      ]
                    ),
                  ],
                ),
              ),
            ),

            const Expanded(
              child: SizedBox(),
            ),

            Visibility(
              visible: _isEventFromUser,
              child: Row(
                mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                children: [
                  Expanded(
                    child: OutlinedButton(
                      child: const Text("Cancel"),
                      onPressed: () {
                        Navigator.pop(context);
                      },
                    ),
                  ),

                  const SizedBox(width: 80),

                  Expanded(
                    child: OutlinedButton(
                      child: const Text("Save"),
                      onPressed: () async {
                        if (_formKey.currentState!.validate()) {
                          String? description;
                          if (_descriptionController.text.isNotEmpty) {
                            description = _descriptionController.text;
                          }

                          if(_currentUser == null) {
                            return;
                          }

                          var event = CalendarEvent(
                            id: initialEvent?.id,
                            user: _currentUser!, // is checked above
                            teamId: _selectedTeam!, // is validated
                            title: _titleController.text,
                            description: description,
                            start: _startTime,
                            end: _endTime,
                          );

                          CalendarEvent savedEvent;
                          if (initialEvent == null) {
                            try {
                              savedEvent = await calendarEventService.add(event);
                            } on ServerException catch (e, s) {
                              loggingService.error(e.toString(), exception: e, trace: s);
                              return;
                            }
                          } 
                          else {
                            try {
                              savedEvent = await calendarEventService.add(event);
                            } on ServerException catch (e, s) {
                              loggingService.error(e.toString(), exception: e, trace: s);
                              return;
                            }
                          }

                          if (context.mounted) {
                            Navigator.pop(context, savedEvent);
                          }
                        }
                      }
                    ),
                  ),
                ],
              ),
            ),

            
          ],
        ),
      ),
    );
  }
}
