import 'dart:async';

import 'package:car_app/exception/permission.dart';
import 'package:car_app/model/point.dart';
import 'package:car_app/model/route.dart';
import 'package:car_app/service/interface/authentication.dart';
import 'package:car_app/service/interface/logging.dart';
import 'package:car_app/service/interface/team.dart';
import 'package:car_app/service/interface/route.dart';
import 'package:car_app/service/location.dart';
import 'package:flutter/foundation.dart';

enum TrackingState {
  starting,
  started,
  stopping,
  stopped,
}

extension TrackingButtonText on TrackingState {
  String get text {
    switch (this) {
      case TrackingState.starting:
        return "Starting Tracking";
      case TrackingState.started:
        return "Tracking";
      case TrackingState.stopping:
        return "Stopping Tracking";
      case TrackingState.stopped:
        return "Start Tracking";
      default:
        return "Start tracking";
    }
  }
}

class TrackingViewModel extends ChangeNotifier {
  final AuthenticationService authenticationService;
  final RouteService routeService;
  final TeamService teamService;
  final LocationService _locationService = LocationService();
  final LoggingService loggingService;
  final List<Point> _points = [];

  late StreamSubscription <Point> _locationSubscription;

  String? _teamId;

  TrackingState _trackingState = TrackingState.stopped;
  String get trackingButtonText => _trackingState.text;

  TrackingViewModel({required this.authenticationService, required this.routeService, required this.teamService, required this.loggingService});

  void locationServiceToForeground() {
    _locationService.setMode(Mode.foreground);
  }

  void locationServiceToBackground() {
    _locationService.setMode(Mode.background);
  }

  void setTeamId(String teamId) {
    _teamId = teamId;
  }

  void toggleTracking(void Function() beforeStart, void Function(Point) onUpdate, void Function(String) onError, Future<bool?> Function() onStartTeamIdNotSet) {
    if (_trackingState == TrackingState.stopped) {
      _startTrackingChecked(beforeStart, onUpdate, onError, onStartTeamIdNotSet);
    }
    else if (_trackingState == TrackingState.started) {
      _stopTracking();
    }
  }

  void _startTrackingChecked(void Function() beforeStart, void Function(Point) onUpdate, void Function(String) onError, Future<bool?> Function() onTeamIdNotSet) {
    if(_teamId == null) {
      onTeamIdNotSet().then((result) {
        if(result ?? false) {
          _startTracking(beforeStart, onUpdate, onError);
        }
      });
    }
    else {
      _startTracking(beforeStart, onUpdate, onError);
    }
  }

  void _startTracking(void Function() beforeStart, void Function(Point) onUpdate, void Function(String) onError) {
    beforeStart();
    _points.clear();

    _setTrackingState(TrackingState.starting);

    _locationSubscription = _locationService.locationUpdate.listen((p) {
      onUpdate(p);
      _points.add(p);
    });

    _locationService.start().then((_) {
      _setTrackingState(TrackingState.started);
    })
    .catchError(test: (e) => e is PermissionException, (e, s) {
      onError(e.toString());
      _setTrackingState(TrackingState.stopped);
      loggingService.error(e.toString(), exception: e, trace: s);
    });
  }

  void _stopTracking() {
    _setTrackingState(TrackingState.stopping);
    _locationSubscription.cancel();
    _locationService.stop().then((_) async {
      _setTrackingState(TrackingState.stopped);
      await routeService.add(Route(points: _points, teamId: _teamId));
    });
  }

  void _setTrackingState(TrackingState state) {
    _trackingState = state;
    notifyListeners();
  }
}
